/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Win32Thread Class & Sample Application.
 *
 * The Initial Developer of the Original Code is
 * Ehsan Akhgari <me@ehsanakhgari.org>.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "stdafx.h"
#include "win32thread.h"

#include <process.h>
#include <cassert>

class Guard
{
public:
	explicit Guard(CWin32Thread * pThread)
		: m_pThread( pThread )
	{
	}
	~Guard()
	{
		m_pThread->Delete();
	}

	CWin32Thread * operator -> () const
	{
		return m_pThread;
	}

private:
	Guard(Guard const &);
	Guard & operator = (Guard const &);

private:
	CWin32Thread * const m_pThread;
};

namespace
{

	class CWin32ProcThread : public CWin32Thread
	{
	public:
		CWin32ProcThread(CWin32Thread::PFN_WIN32THREADENTRY pfn, LPVOID pv)
			: CWin32Thread(),
			m_pfnThreadEntry( pfn ),
			m_pvParams( pv )
		{
		}

	private:
		virtual UINT Main();

	private:
		CWin32Thread::PFN_WIN32THREADENTRY m_pfnThreadEntry;
		LPVOID m_pvParams;
	};

	UINT CWin32ProcThread::Main()
	{
		assert(GetThreadHandle());
		assert(m_pfnThreadEntry != NULL);

		return m_pfnThreadEntry( m_pvParams );
	}

}

CWin32ThreadInfo::CWin32ThreadInfo(void)
	: m_hThread( NULL ),
	m_dwCreationFlags( 0 ),
	m_ulStackSize( 0 ),
	m_ulID( 0 )
{
	m_SecAttribs.bInheritHandle = FALSE;
	m_SecAttribs.lpSecurityDescriptor = NULL;
	m_SecAttribs.nLength = sizeof(SECURITY_ATTRIBUTES);
}

CWin32ThreadInfo::~CWin32ThreadInfo(void)
{
	if (m_hThread)
	{
		::CloseHandle( m_hThread );
	}
}

CWin32ThreadInfo::CWin32ThreadInfo(CWin32ThreadInfo const & rhs)
	: m_hThread( NULL ),
	m_dwCreationFlags( rhs.GetCreationFlags() ),
	m_ulStackSize( rhs.GetStackSize() ),
	m_ulID( rhs.GetThreadID() )
{
	::DuplicateHandle( ::GetCurrentProcess(),
		rhs.GetThreadHandle(),
		::GetCurrentProcess(),
		&m_hThread,
		0,
		rhs.m_SecAttribs.bInheritHandle,
		DUPLICATE_SAME_ACCESS );
}

CWin32ThreadInfo & CWin32ThreadInfo::operator = (CWin32ThreadInfo const & rhs)
{
	if (m_hThread)
	{
		::CloseHandle( m_hThread );
	}

	m_hThread = rhs.GetThreadHandle();
	m_dwCreationFlags = rhs.GetCreationFlags();
	m_ulStackSize = rhs.GetStackSize();
	m_ulID = rhs.GetThreadID();

	::DuplicateHandle( ::GetCurrentProcess(),
		rhs.GetThreadHandle(),
		::GetCurrentProcess(),
		&m_hThread,
		0,
		rhs.m_SecAttribs.bInheritHandle,
		DUPLICATE_SAME_ACCESS );

	return *this;
}

CWin32Thread::CWin32Thread()
	: CWin32ThreadInfo()
{
}

CWin32Thread::~CWin32Thread()
{
}

void CWin32ThreadInfo::Attach(HANDLE hThread)
{
	assert(m_hThread == NULL);

	m_hThread = hThread;
}

HANDLE CWin32ThreadInfo::Detach()
{
	HANDLE const hOldThread = m_hThread;
	Attach( NULL );
	return hOldThread;
}

bool CWin32Thread::StartThread()
{
	assert(m_hThread == NULL);

	m_hThread = reinterpret_cast< HANDLE > (::_beginthreadex(
		&m_SecAttribs,
		m_ulStackSize,
		CWin32Thread::Win32ThreadProc,
		static_cast< CWin32Thread * > (this),
		m_dwCreationFlags,
		&m_ulID ));

	return m_hThread != NULL;
}

bool CWin32ThreadInfo::SuspendThread()
{
	assert(m_hThread);

	return (::SuspendThread( m_hThread ) != ThreadError);
}

bool CWin32ThreadInfo::ResumeThread()
{
	assert(m_hThread);

	return (::ResumeThread( m_hThread )!= ThreadError);
}

bool CWin32ThreadInfo::TerminateThread(DWORD dwExitCode)
{
	assert(m_hThread);

	return ::TerminateThread( m_hThread, dwExitCode ) != FALSE;
}

bool CWin32Thread::OnInitialize()
{
	assert(m_hThread);

	return true;
}

void CWin32Thread::OnTerminate()
{
	assert(m_hThread);
}

bool CWin32ThreadInfo::IsRunning()
{
	return WAIT_OBJECT_0 == ::WaitForSingleObject( m_hThread, 0 );
}

void CWin32ThreadInfo::WaitForThreadTermination()
{
	assert(m_hThread);

	::WaitForSingleObject( m_hThread, INFINITE );
}

bool CWin32ThreadInfo::GetExitCode(DWORD & dwExitCode) const
{
	assert(m_hThread);

	return ::GetExitCodeThread( m_hThread, &dwExitCode ) != FALSE;
}

int CWin32ThreadInfo::GetPriority() const
{
	assert(m_hThread);

	return ::GetThreadPriority( m_hThread );
}

bool CWin32ThreadInfo::SetPriority(int nPriority)
{
	assert(m_hThread);

	return ::SetThreadPriority( m_hThread, nPriority ) != FALSE;
}

bool CWin32ThreadInfo::GetPriorityBoost(bool & bDisable) const
{
	assert(m_hThread);

	BOOL b = FALSE;
	bool bResult = ::GetThreadPriorityBoost( m_hThread, &b ) != FALSE;
	if (bResult)
	{
		bDisable = (b != FALSE);
	}
	return bResult;
}

bool CWin32ThreadInfo::SetPriorityBoost(bool bDisable)
{
	assert(m_hThread);

	return ::SetThreadPriorityBoost( m_hThread, bDisable ) != FALSE;
}

bool CWin32ThreadInfo::GetTimes(FILETIME & ftCreationTime,
							FILETIME & ftExitTime,
							FILETIME & ftKernelTime,
							FILETIME & ftUserTime) const
{
	assert(m_hThread);

	return ::GetThreadTimes( m_hThread,
		&ftCreationTime,
		&ftExitTime,
		&ftKernelTime,
		&ftUserTime ) != FALSE;
}

bool CWin32ThreadInfo::SetAfinityMask(DWORD_PTR dwAfinityMask)
{
	assert(m_hThread);

	return ::SetThreadAffinityMask( m_hThread, dwAfinityMask ) != FALSE;
}

#if (_WIN32_WINNT >= 0x0400)
bool CWin32ThreadInfo::SetIdealProcessor(DWORD dwProcessor)
{
	assert(m_hThread);

	return ::SetThreadIdealProcessor( m_hThread, dwProcessor ) != FALSE;
}
#endif

HANDLE CWin32ThreadInfo::GetThreadHandle() const
{
	assert(m_hThread);

	return m_hThread;
}

unsigned int CWin32ThreadInfo::GetThreadID() const
{
	assert(m_hThread);

	return m_ulID;
}

void CWin32Thread::Delete()
{
	delete this;
}

UINT __stdcall CWin32Thread::Win32ThreadProc(void * pv)
{
	UINT uiResult = static_cast< UINT > (CWin32ThreadInfo::ThreadError);

	__try
	{
		uiResult = CWin32Thread::Win32ThreadProcHelper( pv );
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
	}

	return uiResult;
}

UINT CWin32Thread::Win32ThreadProcHelper(void * pv)
{
	Guard const pThread( static_cast< CWin32Thread * > (pv) );

	if (!pThread->OnInitialize())
	{
		pThread->OnTerminate();
		::_endthreadex( static_cast< unsigned > (ThreadError) );
		return static_cast< unsigned > (ThreadError);
	}

	UINT const uiResult = pThread->Main();

	pThread->OnTerminate();

	return uiResult;
}

CWin32Thread * CWin32ProcThreadCreator::CreateThread(CWin32Thread::PFN_WIN32THREADENTRY pfnThreadEntry,
													 LPVOID pvParams,
													 int nPriority,
													 UINT nStackSize,
													 DWORD dwCreationFlags,
													 LPSECURITY_ATTRIBUTES pSecAttribs)
{
	assert(pfnThreadEntry);
	assert(pvParams);

	CWin32Thread * pThread = new CWin32ProcThread( pfnThreadEntry, pvParams );
	if (!pThread)
	{
		return NULL;
	}
	else
	{
		return CWin32ProcThreadCreator::InternalCreateThread( pThread,
			nPriority,
			nStackSize,
			dwCreationFlags,
			pSecAttribs );
	}
}

CWin32Thread * CWin32ProcThreadCreator::InternalCreateThread(CWin32Thread * pThread,
															 int nPriority,
															 UINT nStackSize,
															 DWORD dwCreationFlags,
															 LPSECURITY_ATTRIBUTES pSecAttribs)
{
	assert(pThread);

	if (!pThread)
		return NULL;

	if (nStackSize)
	{
		pThread->SetStackSize( nStackSize );
	}

	pThread->SetCreationFlags( dwCreationFlags | CREATE_SUSPENDED );

	if (pSecAttribs)
	{
		pThread->SetInheritHandle( pSecAttribs->bInheritHandle != FALSE );
		pThread->SetSecurityDescriptor( pSecAttribs->lpSecurityDescriptor );
	}

	if (!pThread->StartThread())
	{
		pThread->Delete();
		return NULL;
	}

	if (nPriority != THREAD_PRIORITY_NORMAL)
	{
		pThread->SetPriority( nPriority );
	}

	if (!(dwCreationFlags & CREATE_SUSPENDED))
	{
		pThread->ResumeThread();
	}

	return pThread;
}
