/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Win32Thread Class & Sample Application.
 *
 * The Initial Developer of the Original Code is
 * Ehsan Akhgari <me@ehsanakhgari.org>.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef CWin32Thread_Header_Included_
#define CWin32Thread_Header_Included_

#pragma once

#ifndef _MT
#error Use of this header requires the multithreaded version of the Standard Library.
#endif

// This class wraps a thread handle, and provides convenience member functions
// for working with the thread handle
class CWin32ThreadInfo
{
// Construction/Destruction
public:
	CWin32ThreadInfo(void);
	~CWin32ThreadInfo(void);

// Copy construction/assignment
public:
	CWin32ThreadInfo(CWin32ThreadInfo const & rhs);
	CWin32ThreadInfo & operator = (CWin32ThreadInfo const & rhs);

// Attach/Detach semantics
public:
	void Attach(HANDLE hThread);
	HANDLE Detach(void);

// Convenience member functions
public:
	// Facilities to suspend/resume/end a thread
	bool SuspendThread(void);
	bool ResumeThread(void);
	bool TerminateThread(DWORD dwExitCode = static_cast< DWORD > (ThreadError));

	bool IsRunning(void);                                            // Return true if the thread is running
	void WaitForThreadTermination(void);                             // Waits until the thread terminates

	HANDLE GetThreadHandle(void) const;                              // Retrieve the Win32 thread handle

	unsigned int GetThreadID(void) const;                            // Retrieve the Win32 thread ID

	bool GetExitCode(DWORD & dwExitCode) const;                      // Retrieve the thread's exit code (if it has exited)

	int GetPriority(void) const;                                     // Retrieve the thread's priority

	bool SetPriority(int nPriority);                                 // Set the thread's priority

	bool GetPriorityBoost(bool & bDisable) const;                    // Retrieve the thread's priorty boost

	bool SetPriorityBoost(bool bDisable);                            // Set the thread's priorty boost

	bool GetTimes(FILETIME & ftCreationTime, FILETIME & ftExitTime,
		FILETIME & ftKernelTime, FILETIME & ftUserTime) const;       // Retrieve the amount of time the trhead has spent in vasious modes

	bool SetAfinityMask(DWORD_PTR dwAfinityMask);                    // Set the thread's afinity mask

#if (_WIN32_WINNT >= 0x0400)
	bool SetIdealProcessor(DWORD dwProcessor);                       // Set the thread's ideal CPU index
#endif

	void SetStackSize(unsigned long ulStackSize)                     // Set the thread's stack size (before starting the thread)
	{
		m_ulStackSize = ulStackSize;
	}

	unsigned long GetStackSize(void) const                           // Retrieve the thread's stack size
	{
		return m_ulStackSize;
	}

	void SetCreationFlags(DWORD dwCreationFlags)                     // Set the thread's creation flags (before starting the thread)
	{
		m_dwCreationFlags = dwCreationFlags;
	}

	DWORD GetCreationFlags(void) const                               // Retrieve the thread's creation flags
	{
		return m_dwCreationFlags;
	}

	void SetInheritHandle(bool bInherit)                             // Set the thread's handle inheritation mode (before starting the thread)
	{
		m_SecAttribs.bInheritHandle = bInherit ? TRUE : FALSE;
	}

	void SetSecurityDescriptor(LPVOID pvSecDesc)                     // Set the thread's security descriptor (before starting the thread)
	{
		m_SecAttribs.lpSecurityDescriptor = pvSecDesc;
	}

	static HANDLE GetCurrentThread(void)                             // Retrieve a handle to the currently executing thread
	{
		return ::GetCurrentThread();
	}

	static DWORD GetCurrentThreadId(void)                            // Retrieve the ID of the currently executing thread
	{
		return ::GetCurrentThreadId();
	}

	enum { ThreadError = 0xFFFFFFFF };                               // Error return value constant for a thread

// Implementation details
protected:
	HANDLE m_hThread;
	DWORD m_dwCreationFlags;
	unsigned long m_ulStackSize;
	unsigned int m_ulID;
	SECURITY_ATTRIBUTES m_SecAttribs;
};

// This class has two purposes.  It encapsulates a Win32 thread handle,
// and provides conveniece member functions for manipulating it, and also
// it serves as a base class for custom thread types.
class CWin32Thread : public CWin32ThreadInfo
{
// Helper typedefs
public:
	typedef DWORD (WINAPI * PFN_WIN32THREADENTRY) (LPVOID);

// Construction/Destruction
protected:
	CWin32Thread(void);
	virtual ~CWin32Thread(void);
	// Copy construction/assignment done in base class

// Overridables
protected:
	virtual bool OnInitialize(void);  // Called during thread initialization
	virtual void OnTerminate(void);   // Called during thread termination
	virtual UINT Main(void) = 0;      // The main code for the thread. Return value is the
	                                  // return value of the thread.

// Implementation details
private:
	static UINT __stdcall Win32ThreadProc(void *);
	static UINT Win32ThreadProcHelper(void *);
	void Delete(void);
	bool StartThread(void);

// Implementation details
private:
	using CWin32ThreadInfo::m_hThread;
	using CWin32ThreadInfo::m_dwCreationFlags;
	using CWin32ThreadInfo::m_ulStackSize;
	using CWin32ThreadInfo::m_ulID;
	using CWin32ThreadInfo::m_SecAttribs;

	friend class CWin32ProcThreadCreator;
	friend class Guard;
};

// This class's CreateThread( ) function should be used to create a thread
// which uses a free function of type CWin32Thread::PFN_WIN32THREADENTRY
// as its thread procedure
class CWin32ProcThreadCreator
{
// This class is not directly creatable/copyable
private:
	CWin32ProcThreadCreator();
	CWin32ProcThreadCreator(CWin32ProcThreadCreator const &);
	~CWin32ProcThreadCreator();
	CWin32ProcThreadCreator & operator = (CWin32ProcThreadCreator const &);

// This function creates a Win32 thread based upon a free function, and
// wraps the resulting thread handle in a CWin32Thread handle, and returns
// it as a pointer
public:
	static CWin32Thread * CreateThread(CWin32Thread::PFN_WIN32THREADENTRY pfnThreadEntry,
		LPVOID pvParams,
		int nPriority = THREAD_PRIORITY_NORMAL,
		UINT nStackSize = 0,
		DWORD dwCreationFlags = 0,
		LPSECURITY_ATTRIBUTES pSecAttribs = NULL);

// Implementation details
protected:
	static CWin32Thread * InternalCreateThread(CWin32Thread * pThread,
		int nPriority,
		UINT nStackSize,
		DWORD dwCreationFlags,
		LPSECURITY_ATTRIBUTES pSecAttribs);
};

// This template class's CreateThread( ) function is used to create a
// Win32 thread for which there is a CWin32Thread-derived class specified
// by the ThreadType template parameter
template < class ThreadType >
class CWin32ThreadCreator : private CWin32ProcThreadCreator
{
// This class is not directly creatable/copyable
private:
	CWin32ThreadCreator();
	CWin32ThreadCreator(CWin32ThreadCreator const &);
	~CWin32ThreadCreator();
	CWin32ThreadCreator & operator = (CWin32ThreadCreator const &);

// This function creates the thread using the template class specified
// as the template parameter, and returns a pointer to it
public:
	static ThreadType * CreateThread(int nPriority = THREAD_PRIORITY_NORMAL,
		UINT nStackSize = 0,
		DWORD dwCreationFlags = 0,
		LPSECURITY_ATTRIBUTES pSecAttribs = NULL);
};

// Implementation details
template < class ThreadType >
inline ThreadType * CWin32ThreadCreator< ThreadType >::CreateThread(int nPriority,
																	UINT nStackSize,
																	DWORD dwCreationFlags,
																	LPSECURITY_ATTRIBUTES pSecAttribs)
{
	// If the below line fails to compile, it indicates that the template parameter
	// lacks an appropriate default constructor
	ThreadType * pThread = new ThreadType();

	if (!pThread)
	{
		return NULL;
	}

	// If the below line fails to compile, it indicates that the template parameter
	// does not indicate a valid CWin32Thread-derived thread class
	static_cast< CWin32Thread * > (pThread);

	return
		static_cast< ThreadType * > (CWin32ProcThreadCreator::InternalCreateThread( pThread,
			nPriority,
			nStackSize,
			dwCreationFlags,
			pSecAttribs ));
}

#endif